from spacy.tokens import Doc, Span, Token

from .matcher import Matcher


class PatternLabeler:
    def __init__(self, vocab, validate=None):
        Doc.set_extension("labelings", default=[], force=True)
        Token.set_extension("labels", default=[], force=True)

        self._matcher = Matcher(vocab, validate)

    def add(self, label, patterns, on_match=None):
        """
        Add a labeling-rules to the labeler.
        
        Parameters
        ----------
        label: str
            The label related to patterns.
        patterns: list
            The pattern matching rules.
        on_match: callable, optional
            Callback executed on match, by default None.
        """
        self._matcher.add(label, patterns, on_match=on_match)

    def __call__(self, doc: Doc):
        """
        Collect all labels related to patterns matching tokens.
        Each token have a `labels` extension in which to store them.
        The supplied `Doc` have a `labelings` extension in which 
        all labeled spans are collected. 

        Parameters
        ----------
        doc: Doc
            The document to label over.

        Returns
        -------
        Doc
            The doc after labeling.
        """
        matches = self._matcher(doc)
        for key, start, end in matches:
            label = doc.vocab.strings[key]
            span = Span(doc, start, end, label)
            for token in span:
                if label in token._.labels:
                    continue
                token._.labels.append(label)
            doc._.labelings.append(span)
        return doc



import pytest
import spacy
from spacy.util import get_lang_class


def pytest_addoption(parser):
    parser.addoption("--slow", action="store_true", help="include slow tests")


def pytest_runtest_setup(item):
    def getopt(opt):
        # When using 'pytest --pyargs spacy' to test an installed copy of
        # spacy, pytest skips running our pytest_addoption() hook. Later, when
        # we call getoption(), pytest raises an error, because it doesn't
        # recognize the option we're asking about. To avoid this, we need to
        # pass a default value. We default to False, i.e., we act like all the
        # options weren't given.
        return item.config.getoption("--%s" % opt, False)

    for opt in ["slow"]:
        if opt in item.keywords and not getopt(opt):
            pytest.skip("need --%s option to run" % opt)


@pytest.fixture(scope="session")
def nlp():
    return spacy.load("en_core_web_sm")


@pytest.fixture(scope="session")
def en_tokenizer():
    return get_lang_class("en").Defaults.create_tokenizer()


@pytest.fixture(scope="session")
def en_vocab():
    return get_lang_class("en").Defaults.create_vocab()


@pytest.fixture(scope="session")
def en_parser(en_vocab):
    nlp = get_lang_class("en")(en_vocab)
    return nlp.create_pipe("parser")



import pytest

from spikex.abbreviations import AbbreviationDetector, find_abbreviation


def _get_long_form(nlp, text, short):
    doc = nlp(" ".join([text, short]))
    long_end = text.count(" ") + short.count(" ") + 1
    long_form = doc[0:long_end]
    short_start = long_end + short.count("(")
    short_end = short_start + 1
    short_form = doc[short_start:short_end]
    abbr = find_abbreviation(long_form, short_form)
    if abbr is not None:
        return abbr[0]
    return abbr


@pytest.fixture(scope="module")
def detector(nlp):
    return AbbreviationDetector(nlp)


@pytest.mark.parametrize(
    "short", [("(asa)"), ("ASA"), ("(as-9a)"), ("(AS-9A)"),]
)
def test_acronyms_base(nlp, short):
    text = "this is another short abbreviation"
    long_words = _get_long_form(nlp, text, short).text.split()
    assert long_words == text.split()[-3:]


@pytest.mark.parametrize("short", [("(asa)"), ("ASA"),])
def test_acronyms_with_middle_words(nlp, short):
    text = "this is another short in abbreviation"
    long_words = _get_long_form(nlp, text, short).text.split()
    assert long_words == text.split()[-4:]


@pytest.mark.parametrize("short", [("- (asa)"), ("- ASA"), ("-- (ASA)"),])
def test_acronyms_with_end_no_alnum(nlp, short):
    text = "yet another short abbreviation"
    long_words = _get_long_form(nlp, text, short).text.split()
    assert long_words == text.split()[-3:]


@pytest.mark.parametrize("short", [("(asa)"), ("ASA"), ("-- (ASA)"),])
def test_acronyms_bad_long_form(nlp, short):
    text = "this is not our short abbreviation"
    assert _get_long_form(nlp, text, short) == None


@pytest.mark.parametrize("short", [("(asbrv)"), ("ASBRV"), ("TSO")])
def test_acronyms_bad_short_form(nlp, short):
    text = "this is a short abbreviation"
    assert _get_long_form(nlp, text, short) == None


@pytest.mark.parametrize("short", [("(abbrv)"), ("ABBRV"), ("(abb-9rv)"),])
def test_abbreviations_singleword_long_form(nlp, short):
    text = "abbreviation"
    assert _get_long_form(nlp, text, short).text == text


@pytest.mark.parametrize("short", [("(abbrn)"), ("ABBRN"), ("(abb-9rv)")])
def test_abbreviations_multiword_long_form(nlp, short):
    text = "more words are considered aaaabbreviation"
    long_words = _get_long_form(nlp, text, short).text.split()
    assert long_words == text.split()[-1:]


@pytest.mark.parametrize(
    "short", [("- (abbrv)"), ("- ABBRV"), ("-- (abb-9rv)"),]
)
def test_abbreviations_with_no_alnum_end(nlp, short):
    text = "abbreviation"
    assert _get_long_form(nlp, text, short).text == text


@pytest.mark.parametrize("short", [("(aebbrn)"), ("AEBBRN"),])
def test_abbreviations_bad_short_form(nlp, short):
    text = "abbreviation"
    assert _get_long_form(nlp, text, short) == None


@pytest.mark.parametrize(
    "text, short, long",
    [
        ("this is my abbr (MA)", "MA", "my abbr"),
        ("this is other abbr OA", "OA", "other abbr"),
        ("this is TC (too cool)", "TC", "too cool"),
    ],
)
def test_detection_single(detector, nlp, text, short, long):
    doc = detector(nlp(text))
    assert len(doc._.abbreviations) == 1
    assert doc._.abbreviations[0].text == short
    assert doc._.abbreviations[0]._.long_form.text == long


def test_detection_multiple(detector, nlp):
    text = "this is my abbr (MA) and this is MA (my abbr)"
    doc = detector(nlp(text))
    assert len(doc._.abbreviations) == 2
    for abbr in doc._.abbreviations:
        assert abbr.text == "MA"
        assert abbr._.long_form.text == "my abbr"


@pytest.mark.parametrize(
    "text, short, long",
    [
        ("my abbr is cool, this is my abbr (MA)", "MA", "my abbr"),
        ("too cool (TC) is cool, this is TC", "TC", "too cool"),
    ],
)
def test_detection_with_loners(detector, nlp, text, short, long):
    doc = detector(nlp(text))
    assert len(doc._.abbreviations) == 2
    for abbr in doc._.abbreviations:
        assert abbr.text == short
        assert abbr._.long_form.text == long


@pytest.mark.parametrize("text", [("this is not an abbreviation ABB-9V")])
def test_detection_empty(detector, nlp, text):
    doc = detector(nlp(text))
    assert len(doc._.abbreviations) == 0



from spacy.tokens import Span

from spikex.phrases import PhraseDetector


def test_np_simple(nlp):
    text = "a simple noun phrase and a second noun phrase."
    doc = PhraseDetector(nlp.vocab)(nlp(text))
    assert len(doc._.phrases) == 2
    assert doc._.phrases[0] == ("NP", Span(doc, 0, 4))
    assert doc._.phrases[1] == ("NP", Span(doc, 5, 9))


def test_np_complex(nlp):
    text = "this is the long and unexpectedly complex noun phrase."
    doc = PhraseDetector(nlp.vocab)(nlp(text))
    assert len(doc._.phrases) == 1
    assert doc._.phrases[0] == ("NP", Span(doc, 2, 9))


def test_vp_simple(nlp):
    text = "this was created obviously simple."
    doc = PhraseDetector(nlp.vocab)(nlp(text))
    assert len(doc._.phrases) == 1
    assert doc._.phrases[0] == ("VP", Span(doc, 1, 4))


def test_vp_complex(nlp):
    text = "I have been deeply trying to find."
    doc = PhraseDetector(nlp.vocab)(nlp(text))
    assert len(doc._.phrases) == 1
    assert doc._.phrases[0] == ("VP", Span(doc, 1, 5))



from spacy.tokens import Doc, Span

from .matcher import Matcher

NP_PATTERNS = [
    [
        {"POS": "DET", "OP": "?"},
        {"POS": {"IN": ["ADJ", "ADP", "ADV", "NOUN", "PROPN"]}, "OP": "*",},
        {"POS": {"IN": ["CONJ", "CCONJ"]}, "OP": "?"},
        {"POS": {"IN": ["ADJ", "ADP", "ADV", "NOUN", "PROPN"]}, "OP": "+",},
        {"POS": {"IN": ["NOUN", "PRON", "PROPN"]}},
    ]
]

VP_PATTERNS = [
    [
        {"POS": {"IN": ["ADV", "AUX", "VERB"]}, "OP": "+"},
        {"POS": "VERB", "OP": "+"},
        {"POS": "ADV", "OP": "*"},
    ]
]


class PhraseDetector:
    def __init__(self, vocab):
        Doc.set_extension("phrases", default=[], force=True)
        self._matcher = Matcher(vocab)
        self._matcher.add("NP", NP_PATTERNS)
        self._matcher.add("VP", VP_PATTERNS)

    def __call__(self, doc: Doc):
        phrases = []
        good_start = -1
        for key, start, end in self._matcher(doc):
            if good_start >= end:
                continue
            good_start = end
            span = Span(doc, start, end)
            tag = doc.vocab.strings[key]
            phrases.append((tag, span))
        doc._.phrases = phrases
        return doc



import cProfile
import pstats
from pathlib import Path

import spacy
from spacy.matcher import Matcher as SpacyMatcher
from srsly import read_jsonl
from wasabi import msg

from ..matcher import Matcher


def profile(patterns_path, use_spacy: bool = None):
    sample_doc = doc()
    matcher = (
        SpacyMatcher(sample_doc.vocab)
        if use_spacy
        else Matcher(sample_doc.vocab)
    )
    matcher.add("Profile", patterns(patterns_path))
    cProfile.runctx(
        "matches(matcher, sample_doc)", globals(), locals(), "Profile.prof"
    )
    s = pstats.Stats("Profile.prof")
    msg.divider("Profile stats")
    s.strip_dirs().sort_stats("time").print_stats()


def matches(matcher, doc):
    count = 0
    for _, s, e in matcher(doc):
        count += 1
        # print(doc[s:e], s, e)
    msg.text("Total matches", count)


def patterns(patterns_path):
    return list(
        [
            p["pattern"] if "pattern" in p else p
            for p in read_jsonl(patterns_path)
        ]
    )


def doc():
    return spacy.load("en_core_web_sm")(
        open(Path("resources").joinpath("sample.txt"), "r").read()
    )



import pytest
from spacy.tokens import Span

from spikex.util import idx2i, span_idx2i

_TEXT_SAMPLE = "this is a sample text useful for testing"


@pytest.mark.parametrize(
    "idx, i, slice_at",
    [
        (0, 0, 0),  # first token
        (10, 3, 2),  # central token
        (33, 7, 4),  # last token
        (4, 1, 0),  # space in prev token
        (13, 3, 2),  # middle of a token
        pytest.param(-1, 0, 0, marks=pytest.mark.xfail),  # left out of bounds
        pytest.param(40, 0, 0, marks=pytest.mark.xfail),  # right out of bounds
    ],
)
def test_idx2i(nlp, idx, i, slice_at):
    doc = nlp(_TEXT_SAMPLE)
    doc_idx = idx2i(doc, idx)
    fix_idx = idx - sum((len(t) for t in doc[:slice_at]))
    span = Span(doc, slice_at, len(doc))
    span_idx = idx2i(span, fix_idx)
    tokens = doc[slice_at:]
    tokens_idx = idx2i(tokens, fix_idx)
    fix_i = i - slice_at
    assert doc_idx == i
    assert span_idx == fix_i
    assert tokens_idx == fix_i


@pytest.mark.parametrize(
    "start_idx, end_idx, start_i, end_i, slice_at",
    [
        (0, 4, 0, 1, 0),  # first token
        (0, 7, 0, 2, 0),  # first tokens
        (10, 16, 3, 4, 2),  # central token
        (10, 21, 3, 5, 2),  # central tokens
        (33, 40, 7, 8, 4),  # last token
        (29, 40, 6, 8, 4),  # last tokens
        (4, 10, 1, 3, 0),  # start and end with space
        (13, 19, 3, 5, 2),  # start and end in middle of a token
        pytest.param(
            -1, 1, 0, 0, 0, marks=pytest.mark.xfail
        ),  # left out of bounds
        pytest.param(
            40, 41, 0, 0, 0, marks=pytest.mark.xfail
        ),  # right out of bounds
    ],
)
def test_span_idx2i(nlp, start_idx, end_idx, start_i, end_i, slice_at):
    doc = nlp(_TEXT_SAMPLE)
    doc_bounds = span_idx2i(doc, start_idx, end_idx)
    offset_idx = len(doc[:slice_at].text_with_ws)
    fix_start_idx = start_idx - offset_idx
    fix_end_idx = end_idx - offset_idx
    span = Span(doc, slice_at, len(doc))
    span_bounds = span_idx2i(span, fix_start_idx, fix_end_idx)
    tokens = doc[slice_at:]
    tokens_bounds = span_idx2i(tokens, fix_start_idx, fix_end_idx)
    assert doc_bounds == (start_i, end_i)
    fix_bounds = (start_i - slice_at, end_i - slice_at)
    assert span_bounds == fix_bounds
    assert tokens_bounds == fix_bounds



from spikex.labelers import PatternLabeler


def test_labelings(nlp):
    labeler = PatternLabeler(nlp.vocab)
    # first label
    labeler.add("NO_TEST", [[{"TEXT": "test", "OP": "!"}]])
    doc = labeler(nlp("This is a test"))
    assert len(doc._.labelings) == 3
    for labeling in doc._.labelings:
        assert labeling.label_ == "NO_TEST"
    # second label
    labeler.add("TEST", [[{"TEXT": "test"}]])
    doc = labeler(nlp("This is a test"))
    assert doc._.labelings[-1].label_ == "TEST"



from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "jsonschema",
    "regex",
    "spacy>=2.2.2,<3.0",
    "wasabi",
    "typer",
]

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest"]

setup(
    author="Erre Quadro srl",
    author_email="paolo.arduin@errequadrosrl.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="SPIKEX - SpaCy Pipes for Knowledge Extraction",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    include_package_data=True,
    keywords="spikex",
    name="spikex",
    packages=find_packages(include=["spikex", "spikex.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/erre-quadro/spikex",
    version="0.3.0",
    zip_safe=False,
)



import os
import shutil
from io import open as _open
from pathlib import Path

_HERE_PATH = Path(os.path.dirname(__file__))


def open(filename, mode="r", folder=None):
    path = _HERE_PATH
    if folder is not None:
        path = path.joinpath(folder)
        if not path.exists():
            if mode.startswith("r"):
                raise FileNotFoundError(path)
            path.mkdir()
    return _open(path.joinpath(filename), mode)


def contains(filename, folder=None):
    path = _HERE_PATH
    if folder is not None:
        path = path.joinpath(folder)
        if not path.exists():
            return False
    return path.joinpath(filename).exists()


def delete(filename):
    if not contains(filename):
        return
    path = _HERE_PATH.joinpath(filename)
    shutil.rmtree(path)
    return True


def get(filename, default=None):
    if not contains(filename):
        return default
    return force_get(filename)


def force_get(filename):
    return _HERE_PATH.joinpath(filename)



__author__ = "Erre Quadro srl"
__email__ = "paolo.arduin@errequadrosrl.com"
__license__ = "Apache Software License 2.0"
__summary__ = "SPIKEX - SpaCy Pipes for Knowledge Extraction"
__title__ = "spikex"
__version__ = "0.3.0"
